/***
 * Juliac
 * Copyright (C) 2010-2013 Inria, Univ. Lille 1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@ow2.org
 *
 * Author: Lionel Seinturier
 */

package org.objectweb.fractal.juliac.opt.ultramerge;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;

import org.objectweb.fractal.fraclet.annotations.Attribute;
import org.objectweb.fractal.fraclet.annotations.Requires;
import org.objectweb.fractal.fraclet.extensions.Controller;
import org.objectweb.fractal.fraclet.types.Cardinality;
import org.objectweb.fractal.juliac.Juliac;
import org.objectweb.fractal.juliac.api.JuliacRuntimeException;
import org.objectweb.fractal.juliac.api.SpoonSupportItf;
import org.objectweb.fractal.juliac.conf.JuliacConfig;
import org.objectweb.fractal.juliac.desc.AttributeDesc;
import org.objectweb.fractal.juliac.desc.ComponentDesc;
import org.objectweb.fractal.juliac.spoon.SpoonHelper;
import org.objectweb.fractal.juliac.spoon.SpoonSupportImpl;
import org.objectweb.fractal.juliac.ucf.UnifiedClass;
import org.objectweb.fractal.juliac.ucf.UnifiedMethod;

import spoon.reflect.Factory;
import spoon.reflect.declaration.CtAnnotation;
import spoon.reflect.declaration.CtClass;
import spoon.reflect.declaration.CtField;
import spoon.reflect.factory.AnnotationFactory;
import spoon.reflect.factory.TypeFactory;
import spoon.reflect.reference.CtTypeReference;
import spoon.reflect.visitor.Query;
import spoon.reflect.visitor.filter.AbstractFilter;

/**
 * Source code generator which merges all component business code in a single
 * class. This class handles component business code annotated with the Fraclet
 * framework. 
 * 
 * @author Lionel Seinturier <Lionel.Seinturier@univ-lille1.fr>
 * @since 2.2.4
 */
public class FracletUltraMerge extends AbstractUltraMerge {

    public FracletUltraMerge( Juliac jc ) {
        super(jc);
    }

    // ----------------------------------------------------------------------
    // Abstract methods
    // ----------------------------------------------------------------------
    
    @Override
    protected void mergeabilityCheck( ComponentDesc cdesc )
    throws IllegalComponentCodeException {
        super.mergeabilityCheck(cdesc);
        mergeabilityCheckNoControllerAnnotation();
    }
    
    protected String getRequiredAnnotatedUMFieldName( UMField<?> umfield ) {
        
        CtField<?> ctfield = umfield.getCtField();
        Requires r = ctfield.getAnnotation(Requires.class);
        String name = r.name();
        
        if( name.length() == 0 ) {
            /*
             * No name parameter for the @Requires annotation.
             * Return the field name.
             */
            name = umfield.getSimpleName();
        }

        return name;
    }
    
    protected boolean isRequiredAnnotatedUMField( UMField<?> umfield ) {
        CtField<?> ctfield = umfield.getCtField();
        Requires r = ctfield.getAnnotation(Requires.class);
        boolean b = r != null;
        return b;
    }

    protected boolean
        isRequiredAnnotatedUMFieldCardinalitySingleton( UMField<?> umfield ) {
        
        CtField<?> ctfield = umfield.getCtField();
        Requires r = ctfield.getAnnotation(Requires.class);
        boolean b = r.cardinality() == Cardinality.SINGLETON;
        return b;
    }

    protected boolean
        isRequiredAnnotatedUMFieldCardinalityCollection( UMField<?> umfield ) {
        
        CtField<?> ctfield = umfield.getCtField();
        Requires r = ctfield.getAnnotation(Requires.class);
        boolean b = r.cardinality() == Cardinality.COLLECTION;
        return b;
    }
    
    protected void annotateRequiredCtField(
        AnnotationFactory af, CtField<?> field, String cltItfName ) {
        af.annotate(field,Requires.class,"name",cltItfName);        
    }
    
    protected boolean isRequiredAnnotation( Annotation a ) {
        boolean b = a instanceof Requires;
        return b;
    }

    /**
     * Return the setter method for the specified attribute name.
     */
    protected UMMethod<?> getAttributeSetterUMMethod(
        UMClass<?> umclass, AttributeDesc adesc ) {
        
    	JuliacConfig jconf = jc.getJuliacConfig();
        SpoonSupportImpl spoon =
        	(SpoonSupportImpl) jconf.lookupUnique(SpoonSupportItf.class);
        Factory factory = spoon.getSpoonFactory();
        TypeFactory tf = factory.Type();

        UnifiedClass returnType = getAttributeType(adesc);
        CtTypeReference<?> tref = tf.createReference(returnType.getName());
        String name = adesc.getName();
        String setterMethodName =
            "set"+name.substring(0,1).toUpperCase()+name.substring(1);
        
        List<CtTypeReference<?>> paramTypes =
            new ArrayList<CtTypeReference<?>>();
        paramTypes.add(tref);
        UMMethod<?> ummethod = umclass.getUMMethod(setterMethodName,paramTypes);
        
        return ummethod;
    }
    
    protected boolean isAttributeAnnotatedCtField( CtField<?> ctfield ) {
        Attribute r = ctfield.getAnnotation(Attribute.class);
        boolean b = r != null;
        return b;
    }

    protected String getAttributeAnnotatedCtFieldName( CtField<?> ctfield ) {
        Attribute r = ctfield.getAnnotation(Attribute.class);
        String name = r.name();
        return name;
    }
    
    protected CtTypeReference<?> getCollectionFieldType( UMField<?> umfield ) {
        CtField<?> ctfield = umfield.getCtField();
        CtTypeReference<?> fieldType = ctfield.getType();
        List<CtTypeReference<?>> genericTypes =
            fieldType.getActualTypeArguments();
        if( genericTypes==null || genericTypes.size()!=2 ) {
            String msg =
                SpoonHelper.toEclipseClickableString(ctfield)+
                " expected to be of type Map<K,V>";
            throw new JuliacRuntimeException(msg);
        }
        CtTypeReference<?> refType = genericTypes.get(1);
        return refType;
    }
    
    /**
     * Return the class for variables which store references for collection
     * interfaces.
     */
    protected Class<?> getCollectionClass() {
        return HashMap.class;
    }

        
    // ----------------------------------------------------------------------
    // Checks
    // ----------------------------------------------------------------------
    
    /**
     * Check that the {@link CtClass}es of the architecture do not contain
     * @{@link Controller} annotated fields.
     */
    private void mergeabilityCheckNoControllerAnnotation()
    throws IllegalComponentCodeException {
        
        Collection<UMClass<?>> umclasses = repo.getAllUMClasses();
        for (UMClass<?> umclass : umclasses) {
            CtClass<?> src = umclass.getCtClass();
            mergeabilityCheckNoControllerAnnotation(src);
        }
    }
    
    /**
     * Check that the specified {@link CtClass} does not contain @{@link
     * Controller} annotated fields.
     * 
     * @param src  the {@link CtClass} to be checked against mergeability
     * @throws IllegalComponentCodeException  if this is not the case
     */
    private void mergeabilityCheckNoControllerAnnotation( CtClass<?> src )
    throws IllegalComponentCodeException {
        
        List<CtAnnotation<?>> annots =
            Query.getElements(
                src,
                new AbstractFilter<CtAnnotation<?>>(CtAnnotation.class) {
                    public boolean matches( CtAnnotation<?> ctannot ) {
                        Annotation annot = ctannot.getActualAnnotation();
                        return ( annot instanceof Controller );
                    }
                }
            );
        
        if( annots.size() != 0 ) {
            StringBuffer msg = new StringBuffer();
            msg.append("@Controller annotations at the following source ");
            msg.append("position prevent the code from being merged");
            for (CtAnnotation<?> annot : annots) {
                msg.append("\n");
                String pos = toPositionString(annot);
                msg.append(pos);
            }
            throw new IllegalComponentCodeException(msg.toString());
        }
    }    

    
    // ----------------------------------------------------------------------
    // Implementation specific
    // ----------------------------------------------------------------------
    
    /**
     * Return the type of the specified attribute descriptor.
     */
    private UnifiedClass getAttributeType( AttributeDesc adesc )
    throws JuliacRuntimeException {
        
        String ac = adesc.getAttributeController();
        String name = adesc.getName();
        
        UnifiedClass uc = jc.create(ac);
        try {
            String methName =
                "get" + name.substring(0,1).toUpperCase() +
                name.substring(1);
            UnifiedMethod um = uc.getMethod(methName);
            UnifiedClass returnType = um.getReturnType();
            return returnType;
        }
        catch (NoSuchMethodException e) {
            // Should not occur
            throw new JuliacRuntimeException(e);
        }
    }    
}
